.setcpu "65816"
.smart

;Included Files

.include "Header.asm"

;Define Stuff

HVBJOY=		$4212
JOY0=		$4218
JOY1=		$421a
JOY2=		$421c
JOY3=		$421e
JOYSER0=	$4016
JOYSER1=	$4017
WRIO=		$4201

;Button Numbers
RButton=	16
LButton= 	32
XButton= 	64
AButton= 	128

RightDPad=	1
LeftDPad=	2
DownDPad=	4
UpDPad=		8
StartButton=	16
SelectButton=	32
YButton=	64
BButton=	128

WordRightDPad=		256
WordLeftDPad=		512
WordDownDPad=		1024
WordUpDPad=		2048
WordStartButton=	4096
WordSelectButton=	8192
WordYButton=		16384
WordBButton=		32768

;Tile Request Table Format
VramAddress=	0
TileAddress=	2
BankNumber=	4	

;=========================================================================================
;=========================================================================================

.struct ObjectSlot
Identity	.byte 3
XPosition	.byte 3
YPosition	.byte 3
CurrentFrame	.byte 2
RequestedFrame	.byte 2
VramOffset	.byte 2
MetaspriteTableOffset	.byte 2
FlipBits	.byte 2
.endstruct

;=========================================================================================
.segment "ZEROPAGE"
;=========================================================================================

Joy1Data:		.res 2
Joy2Data:		.res 2
Joy1Press:		.res 2
Joy2Press:		.res 2
Temporary1:		.res 2
SpriteCount:		.res 2
ObjectTableOffset:	.res 2
NewObjectRequest:	.res 3
ColorCounter:		.res 2
LongJumpLocation:	.res 3
HFlipMask:		.res 2
VFlipMask:		.res 2
MetaspriteCount:	.res 2
MetaspriteAttributes:	.res 2
TileRequestCounter16x16:	.res 2
MetaspriteBaseTileNumber:	.res 2
MetaspriteVramOffset:		.res 2
NumberOfVramSlotsRequested:	.res 2
TilesInFrame:		.res 2
VramAddressOfFrame:	.res 2
BankByteOfFrame:	.res 2

;SNES RAM area, $0200-1fff
;See snes.cfg

;=========================================================================================
.segment "BSS"
;=========================================================================================

SpriteBuf1:		.res $200	;Sprite buffer for OAM DMA, 512 ($200) bytes
SpriteBuf1Size=		*-SpriteBuf1
SpriteBuf2:    		.res $20	;Sprite buffer for OAM DMA, 32  ($20) bytes
SpriteBuf2Size=		*-SpriteBuf2
SpriteBufSize= 		SpriteBuf1Size+SpriteBuf2Size
SpriteBuf3:		.res $200	;Sprite buffer for OAM DMA, 512  ($200) bytes
SpriteBuf3Size=		*-SpriteBuf3
ObjectSlotSize=		.sizeof(ObjectSlot)
ObjectSlotNumber=	16
ObjectTable:		.res ObjectSlotSize*ObjectSlotNumber
ObjectTableSize=	 *-ObjectTable
TileRequestSlotSize=	6
TileRequestSlotNumber=	128
TileRequestTable:	.res TileRequestSlotSize*TileRequestSlotNumber
VramSlotSize=		2
VramSlotNumber=		128
VramLinkedListTable:	.res VramSlotSize*VramSlotNumber

AnimationFrameSlotUsageTable:	.res 4
AnimationFrameLinkedListTable:	.res 4

;=========================================================================================
.segment "BSS7E"
;=========================================================================================

;Assemble-time assertion; SNES DMA transfer can only handle 65536 bytes.
.assert (SpriteBufSize < $10000), error, "SpriteBufs area too big for SNES DMA transfer"

;=========================================================================================
.segment "CODE"
;=========================================================================================

.proc Main

  sei		;Disable interrupts
  clc		;Clear carry, used by XCE
  xce		;Set 65816 native mode
  jmp :+	;Needed to set K (bank of PC) properly if MODE 21 is ever used;
		;see official SNES developers docs, "Programming Cautions"

: cld		;Disable decimal mode
  phk		;Push K (PC bank)
  plb		;Pull B (data bank, i.e. data bank now equals PC bank)

  rep #$30	;A=16, X/Y=16

;Note: this should correlate with ZEROPAGE in lorom256k.cfg
  lda #$0000
  tcd		;Set D = $0000 (direct page)

;Note: this should correlate with the top of BSS in lorom256k.cfg
  ldx #$1fff
  txs		;Set X = $1fff (stack pointer)

;Register initialisation values, per official Nintendo documentation
;Note: if you are looking at this code and thinking "how can I optimise"
;this?", then you are doing nothing useful with your time.  This code is
;called ONCE on reset or power-on, it does not need to be optimised.  I
;(koitsu) find it remarkable how many people have tried to "optimise"
;this routine.  STOP IT.

  sep #$20	;A=8
  lda #$80
  sta $2100
  stz $2101
  stz $2102
  stz $2103
  stz $2104
  stz $2105
  stz $2106
  stz $2107
  stz $2108
  stz $2109
  stz $210A
  stz $210B
  stz $210C
  stz $210D
  stz $210D
  stz $210E
  stz $210E
  stz $210F
  stz $210F
  stz $2110
  stz $2110
  stz $2111
  stz $2111
  stz $2112
  stz $2112
  stz $2113               
  stz $2113               
  stz $2114
  stz $2114
  lda #$80
  sta $2115
  stz $2116
  stz $2117
  stz $211A
  stz $211B
  lda #$01
  sta $211B
  stz $211C
  stz $211C
  stz $211D
  stz $211D
  stz $211E       
  lda #$01
  sta $211E
  stz $211F
  stz $211F
  stz $2120
  stz $2120
  stz $2121
  stz $2123
  stz $2124
  stz $2125
  stz $2126
  stz $2127
  stz $2128
  stz $2129
  stz $212A
  stz $212B
  stz $212C
  stz $212D
  stz $212E
  stz $212F
  stz $4200
  lda #$FF
  sta $4201
  stz $4202
  stz $4203
  stz $4204
  stz $4205
  stz $4206
  stz $4207
  stz $4208
  stz $4209
  stz $420A
  stz $420B
  stz $420C
  stz $420D

  jsr clear_vram	;Zero VRAM. Requires A=8, X/Y=16
  jsr clear_cgram	;Zero palette. Requires A=8, X/Y=16
  
;clear sprite tables

  stz $2102	;sprites initialized to be off the screen, palette 0, character 0
  stz $2103
  ldx #$0080
  lda #$F0

loop08:
  sta $2104	;set X = 240
  sta $2104	;set Y = 240
  sta $2104	;set character = $00
  sta $2104	;set priority=0, no flips
  dex
  bne loop08
  ldx #$0020

loop09:
  stz $2104		;set size bit=0, x MSB = 0
  dex
  bne loop09

;clear SNES RAM

  stz $2181	;set WRAM address to $000000
  stz $2182
  stz $2183

  ldx #$8008
  stx $4300		;Set DMA mode to fixed source, BYTE to $2180
  ldx #.LOWORD(zero_fill_byte)
  stx $4302		;Set source offset
  lda #.BANKBYTE(zero_fill_byte)
  sta $4304		;Set source bank
  ldx #$0000
  stx $4305		;Set transfer size to 64KBytes (65536 bytes)
  lda #$01
  sta $420B		;Initiate transfer

  lda #$01		;now zero the next 64KB (i.e. 128KB total)
  sta $420B		;Initiate transfer

  rep #$30	;A=16, X/Y=16
  sep #$20	;A=8

  jsr initialize_oam
  jsr initialize_controllers
  jsr setup_video		;setup the video modes and such, then turn on the screen

;Enable NMI (when interrupts get re-enabled) and joypad reading
  sep #$20	;A=8
  lda #$81
  sta $4200

  cli			;Restore interrupts -- we're all ready to go!

  rep #$30	;A=16, X/Y=16
  lda #.LOWORD(dummy)
  sta NewObjectRequest
  sep #$20	;A=8
  lda #.BANKBYTE(dummy)
  sta NewObjectRequest+2

  rep #$30	;A=16, X/Y=16

  lda #TestMetasprite
  sta ObjectTable+ObjectSlot::MetaspriteTableOffset

  jsr object_spawner

;=========================================================================================

infinite_loop:
  wai			;Wait for interrupts to finish (NMI/VBlank)

  rep #$30	;A=16, X/Y=16

  jsr object_identifier

  rep #$30	;A=16, X/Y=16
  inc ColorCounter
  sep #$20	;A=8
  lda #$00
  sta $2121
  lda ColorCounter
  sta $2122
  lda ColorCounter+1
  sta $2122

  jmp infinite_loop	;Do this forever
.endproc

;=========================================================================================
;=========================================================================================

.proc VBlank
  rep #$30	;A=16, X/Y=16
  pha
  phx
  phy
  phd
  phb
  php
  lda #$0000
  tcd

  jsr tile_uploader

;transfer sprite data

  sep #$20	;A=8
  stz $2102	;set OAM address to 0
  stz $2103	;...and no priority rotation

  stz $4300			;CPU -> PPU, auto-increment, write once
  lda #$04
  sta $4301			;DMA will write to register $2104 (OAM data write)
  ldy #.LOWORD(SpriteBuf1)
  sty $4302			;source offset of SpriteBuf1
  lda #.BANKBYTE(SpriteBuf1)
  sta $4304			;bank address of SpriteBuf1
  ldy #SpriteBufSize
  sty $4305			;transfer SpriteBufSize bytes (pre-calculated number)
  lda #$01
  sta $420B		;start DMA transfer

  jsr get_input	;update joypad data

  lda $4210	;clear NMI Flag

  plp
  plb
  pld
  ply
  plx
  pla
  rti
.endproc

;=========================================================================================
;=========================================================================================

.proc clear_vram
  rep #$30	;A=16, X/Y=16
  sep #$20	;A=8
  lda #$80
  sta $2115		;Set VRAM port to word access
  ldx #$1809
  stx $4300		;Set DMA mode to fixed source, WORD to $2118/9
  ldx #$0000
  stx $2116			;Set VRAM port address to $0000
  ldx #.LOWORD(zero_fill_byte)
  stx $4302			;Set source address to $xx:0000
  lda #.BANKBYTE(zero_fill_byte)
  sta $4304			;Set source bank
  ldx #$0000
  stx $4305		;Set transfer size to 65536 bytes
  lda #$01
  sta $420B		;Initiate transfer
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc clear_cgram
  rep #$30	;A=16, X/Y=16
  sep #$20	;A=8
  stz $2121
  ldx #$0100

clear_palette_loop:
  stz $2122
  stz $2122
  dex
  bne clear_palette_loop
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc initialize_oam
  php	
  rep #$30

  lda #$01
  ldx #0

initialize_oam_loop:
  sta SpriteBuf3+1,x
  inx
  inx
  inx
  inx
  cpx #$0200
  bne initialize_oam_loop
  plp
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc initialize_controllers
  php	

  rep	#$10	;8 bit mem/A, 16 bit X/Y
  sep	#$20

  lda #$C0	;have the automatic read of the SNES read the first pair of JoyPads
  sta WRIO

  ldx #$0000
  stx Joy1Press
  stx Joy2Press

;in ZSNES, the auto-read doesn't work unless NMI is enabled... is this correct?
;
; lda $4212
; lda $4212
;
;w11:
; lda $4212
; and #$01
; beq _w11	;Wait till automatic JoyPort read starts
;
; jsr GetInput	;fill the variables with real Joypad data
;
;w12:
; lda $4212
; and #$01
; beq _w12	;Wait till automatic JoyPort read starts
;
; jsr get_input	;fill the variables with real Joypad data

  plp
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc get_input
  php
  sep #$20	;A=8

  lda #$01
: bit $4212
  bne :-		;Wait till automatic JoyPort read is complete

  rep #$30   ;A=16, X/Y=16

;Get Joypads 1,2

  lda JOY0		;get JoyPad1
  tax 			
  eor Joy1Data		;A = A xor JoyState = (changes in joy state)
  stx Joy1Data		;update JoyState
  ora Joy1Press		;A = (joy changes) or (buttons pressed)
  and Joy1Data		;A = ((joy changes) or (buttons pressed)) and (current joy state)  
  sta Joy1Press		;store A = (buttons pressed since last clearing reg) and (button is still down)

  lda JOY1		;get JoyPad2
  tax 			
  eor Joy2Data		;A = A xor JoyState = (changes in joy state)
  stx Joy2Data		;update JoyState
  ora Joy2Press		;A = (joy changes) or (buttons pressed)
  and Joy2Data		;A = ((joy changes) or (buttons pressed)) and (current joy state)  
  sta Joy2Press		;store A = (buttons pressed since last clearing reg) and (button is still down)

;Make sure Joypads 1,2 are valid

  sep #$30	;A=8, X/Y=8
  lda JOYSER0
  eor #$01
  and #$01		;A = -bit0 of JoySer0
  ora Joy1Data
  sta Joy1Data		;joy state = (joy state) or A.... so bit0 of Joy1State = 0 only if it is a valid joypad

  lda JOYSER1
  eor #$01
  and #$01		;A = -bit0 of JoySer1
  ora Joy2Data		
  sta Joy2Data		;joy state = (joy state) or A.... so bit0 of Joy1State = 0 only if it is a valid joypad

;Change all invalid joypads to have a state of no button presses

  rep #$30	;A=16, X/Y=16

  ldx #$0001
  lda #$000f

  bit Joy1Data		;A = joy state, if any of the bottom 4 bits are on... either nothing is plugged
  beq :+		;into the joy port, or it is not a joypad
  stx Joy1Data		;if it is not a valid joypad put $0001 into the 2 joy state variables
  stz Joy1Press

: bit Joy2Data		;A = joy state, if any of the bottom 4 bits are on... either nothing is plugged
  beq done		;into the joy port, or it is not a joypad
  stx Joy2Data		;if it is not a valid joypad put $0001 into the 2 joy state variables
  stz Joy2Press

done:
  plp
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc setup_video
  php
  rep #$30	;A=16, X/Y=16
  sep #$20	;A=8

  stz $211b
  lda #$01
  sta $211b
  stz $211c
  stz $211c
  stz $211d
  stz $211d
  stz $211e
  sta $211e
  lda #$80
  sta $211f
  stz $211f
  sta $2120
  stz $2120

  lda #%00000000	;Sprites 8x8 or 16x16, character data at $0000 (word address)
  sta $2101         

  lda #$01		;Set video mode 1, 16x16 tiles (16 color BG1, 16 color BG2, 4 color BG3)
  sta $2105         

  lda #$70		;Set BG1's Tile Map VRAM offset to $7800 (word address)
  sta $2107		;and the Tile Map size to 32 tiles x 32 tiles

  lda #$02		;Set BG1's Character VRAM offset to $4000 (word address)
  sta $210B

  lda #$10		;Turn on Sprites
  sta $212C

  lda #$0F		;Turn on screen, full brightness
  sta $2100

  plp
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc process_oam
  php
  rep #$10
  sep #$20

  ldx SpriteCount
  lda #$01

clear_oam_loop:
  cpx #$0200
  beq finished_clearing_oam_loop
  sta SpriteBuf1,x
  stz SpriteBuf1+1,x
  stz SpriteBuf1+2,x
  stz SpriteBuf1+3,x
  sta SpriteBuf3+1,x		;sprite 3, 9th x bit
  inx
  inx
  inx
  inx
  bra clear_oam_loop

finished_clearing_oam_loop:
  ldy #$0000
  sty SpriteCount

build_hioam_loop:
  lda SpriteBuf3+14,y		;sprite 3, size bit
  asl
  ora SpriteBuf3+13,y		;sprite 3, 9th x bit
  asl
  ora SpriteBuf3+10,y		;sprite 2, size bit
  asl
  ora SpriteBuf3+9,y		;sprite 2, 9th x bit
  asl
  ora SpriteBuf3+6,y		;sprite 1, size bit
  asl
  ora SpriteBuf3+5,y		;sprite 1, 9th x bit
  asl
  ora SpriteBuf3+2,y		;sprite 0, size bit
  asl
  ora SpriteBuf3+1,y		;sprite 0, 9th x bit
  sta SpriteBuf1,x
  rep #$20
  tya
  clc
  adc #$0010
  tay
  sep #$20
  inx
  cpx #$0220
  bne build_hioam_loop

  plp
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc object_spawner
  rep #$30    ; A=16, X/Y=16
  ldx #$0000

object_spawner_loop:
  lda a:ObjectTable+ObjectSlot::Identity,x
  bne continue_object_spawner_loop
  lda a:ObjectTable+ObjectSlot::Identity+1,x
  beq object_spawner_loop_done
continue_object_spawner_loop:
  txa
  clc
  adc #ObjectSlotSize
  tax
  cpx #ObjectTableSize
  bne object_spawner_loop

object_spawner_loop_done:
  lda a:NewObjectRequest
  sta a:ObjectTable+ObjectSlot::Identity,x
  lda a:NewObjectRequest+1
  sta a:ObjectTable+ObjectSlot::Identity+1,x
  rts

.endproc

;=========================================================================================
;=========================================================================================

.proc object_identifier
  rep #$30	; A=16, X/Y=16
  lda #ObjectTable
  sta ObjectTableOffset
  tcd

object_identifier_loop:
  lda ObjectSlot::Identity
  cmp #$0001
  bcs continue_object_identifier_loop
  lda ObjectSlot::Identity+1
  beq next_object
  lda ObjectSlot::Identity
continue_object_identifier_loop:
  sta a:LongJumpLocation
  lda ObjectSlot::Identity+1
  sta a:LongJumpLocation+1
  jml [LongJumpLocation]
next_object:
  rep #$30	; A=16, X/Y=16
  lda a:ObjectTableOffset		;says how many objects have been identified
  clc
  adc #ObjectSlotSize			;add 24 to look at the next object
  sta a:ObjectTableOffset		;store the result for the next time we go through the loop
  tcd					;transfer the accumulator to y for "cpy #8192" and "lda ObjectTable,y"
  cmp #ObjectTableSize+ObjectTable	;sees if all objects have been identified (each object is 48 bytes)
  bne object_identifier_loop		;if so, quit searching
  jsr vram_engine
  jsr metasprite_handler
  lda #$0000
  tcd
  jsr process_oam
  rts

.endproc

;=========================================================================================
;=========================================================================================

.proc dummy
  rep #$30	; A=16, X/Y=16

up:
  lda a:Joy1Data
  and #WordUpDPad
  beq down
  dec ObjectSlot::YPosition
  lda ObjectSlot::FlipBits
  and #$7FFF
  sta ObjectSlot::FlipBits

down:
  lda a:Joy1Data
  and #WordDownDPad
  beq left
  inc ObjectSlot::YPosition
  lda #$7FFF
  lda ObjectSlot::FlipBits
  ora #$8000
  sta ObjectSlot::FlipBits

left:
  lda a:Joy1Data
  and #WordLeftDPad
  beq right
  dec ObjectSlot::XPosition
  lda ObjectSlot::FlipBits
  and #$BFFF
  sta ObjectSlot::FlipBits

right:
  lda a:Joy1Data
  and #WordRightDPad
  beq a_button
  inc ObjectSlot::XPosition
  lda ObjectSlot::FlipBits
  ora #$4000
  sta ObjectSlot::FlipBits

a_button:
  lda a:Joy1Data
  and #AButton
  beq b_button
  lda #$0002
  sta ObjectSlot::RequestedFrame

b_button:
  lda a:Joy1Data
  and #WordBButton
  beq dummy_done
  lda #$0004
  sta ObjectSlot::RequestedFrame

dummy_done:
  jmp object_identifier::next_object
.endproc

;=========================================================================================
;=========================================================================================

.proc vram_engine
  rep #$30	;A=16, X/Y=16
  lda #ObjectTable
  tcd
  ldy #$0002

vram_engine_loop:
  ldx ObjectSlot::CurrentFrame
  beq no_current_frame
  cpx ObjectSlot::RequestedFrame
  bne delete_vram

next_object:
  lda ObjectSlot::RequestedFrame
  sta ObjectSlot::CurrentFrame
  inx
  inx
  tdc
  clc
  adc #ObjectSlotSize
  cmp #ObjectTable+ObjectTableSize
  bcs vram_engine_done
  tcd
  bra vram_engine_loop

vram_engine_done:
  rts

;=========================================================================================

delete_vram:
  stz ObjectSlot::CurrentFrame
  dec a:AnimationFrameSlotUsageTable-2,x
  lda a:AnimationFrameSlotUsageTable-2,x
  bne no_current_frame
  lda a:TilesInFrameTable-2,x
  sta a:TilesInFrame
  lda a:AnimationFrameLinkedListTable-2,x
  tay

delete_vram_loop:
  lda a:VramLinkedListTable-2,y
  pha
  lda #$0000
  sta a:VramLinkedListTable-2,y
  ply
  dec a:TilesInFrame
  bne delete_vram_loop

no_current_frame:
  ldx ObjectSlot::RequestedFrame
  beq next_object
  lda a:AnimationFrameSlotUsageTable-2,x
  beq find_vram
  inc a:AnimationFrameSlotUsageTable-2,x
  lda a:AnimationFrameLinkedListTable-2,x
  sta ObjectSlot::VramOffset
  lda ObjectSlot::RequestedFrame
  sta ObjectSlot::CurrentFrame
  bra next_object

;=========================================================================================

find_vram:
  ldy #$0002
  inc a:AnimationFrameSlotUsageTable-2,x
  lda a:TilesInFrameTable-2,x
  sta a:TilesInFrame
  lda a:VramAddressFrameTable-2,x
  sta a:VramAddressOfFrame
  lda a:VramBankByteFrameTable-2,x
  sta a:BankByteOfFrame

find_vram_loop_1:
  lda a:VramLinkedListTable-2,y
  beq open_slot_found_1
  iny
  iny
  cpy #$0102
  bcc find_vram_loop_1
  bra next_object

open_slot_found_1:
  tya
  sta a:AnimationFrameLinkedListTable-2,x
  sta ObjectSlot::VramOffset

  tyx
  phy
  ldy a:TileRequestCounter16x16
  lda a:VramAddressToTransferAddressTable-2,x
  sta a:TileRequestTable+VramAddress,y
  lda a:VramAddressOfFrame
  sta a:TileRequestTable+TileAddress,y
  lda a:BankByteOfFrame
  sta a:TileRequestTable+BankNumber,y
  lda a:VramAddressOfFrame
  clc
  adc #$0080
  sta a:VramAddressOfFrame
  lda a:TileRequestCounter16x16
  clc
  adc #$0006
  sta a:TileRequestCounter16x16
  ply

  dec a:TilesInFrame
  beq jump_to_next_object
  iny
  iny
  cpy #$0102
  bcc find_vram_loop_2
jump_to_next_object:
  brl next_object

find_vram_loop_2:
  lda a:VramLinkedListTable-2,y
  beq open_slot_found_2
  iny
  iny
  cpy #$0102
  bcc find_vram_loop_2
  brl next_object

open_slot_found_2:
  tya
  sta a:VramLinkedListTable-2,x

  tyx
  phy
  ldy a:TileRequestCounter16x16
  lda a:VramAddressToTransferAddressTable-2,x
  sta a:TileRequestTable+VramAddress,y
  lda a:VramAddressOfFrame
  sta a:TileRequestTable+TileAddress,y
  lda a:BankByteOfFrame
  sta a:TileRequestTable+BankNumber,y
  lda a:VramAddressOfFrame
  clc
  adc #$0080
  sta a:VramAddressOfFrame
  lda a:TileRequestCounter16x16
  clc
  adc #$0006
  sta a:TileRequestCounter16x16
  ply

  dec a:TilesInFrame
  beq last_slot
  iny
  iny
  cpy #$0102
  bcc find_vram_loop_2
  brl next_object

last_slot:
  lda #$FFFF
  sta a:VramLinkedListTable-2,x
  brl next_object
.endproc

;=========================================================================================
;=========================================================================================

.proc metasprite_handler
  rep #$30	;A=16, X/Y=16
  lda #ObjectTable
  tcd
  ldy #$0000
  bra continue_metasprite_finder

metasprite_finder:
  tdc
  clc
  adc #ObjectSlotSize
  tcd
  cmp #ObjectTable+ObjectTableSize
  bne continue_metasprite_finder
  sty a:SpriteCount
  rts

continue_metasprite_finder:
  ldx ObjectSlot::MetaspriteTableOffset
  beq metasprite_finder
  lda a:$0000,x
  cmp #$0001
  bne multiple_metasprite
  brl single_metasprite

multiple_metasprite:
  sta a:MetaspriteCount
  phx
  lda ObjectSlot::VramOffset
  sta a:MetaspriteVramOffset
  tax
  lda a:VramAdressToTileNumberTable-2,x
  sta a:MetaspriteBaseTileNumber
  plx

horizontal_flip_check:
  stz a:HFlipMask
  lda ObjectSlot::FlipBits
  bit #$4000
  beq vertical_flip_check
  lda #$FFFF
  sta a:HFlipMask

vertical_flip_check:
  stz a:VFlipMask
  lda ObjectSlot::FlipBits
  bit #$8000
  beq metasprite_loop
  lda #$FFFF
  sta a:VFlipMask

metasprite_loop:
  cpy #$0200			;sees if all 128 sprites are used up
  bne continue_metasprite_loop1
  brl metasprite_finder

continue_metasprite_loop1:
  lda a:MetaspriteCount
  bne continue_metasprite_loop2
  brl metasprite_finder

continue_metasprite_loop2:
  lda a:$0006,x
  and #$0001
  bne sprite_not_flipped_horizontally
  sta a:SpriteBuf3+2,y		;sprite size
  lda ObjectSlot::FlipBits
  and #$4000
  beq sprite_not_flipped_horizontally
  lda a:$0002,x
  clc
  eor a:HFlipMask
  adc ObjectSlot::XPosition
  clc
  adc #$0008
  cmp #256
  bcc sprite_x_not_out_of_bounds
  cmp #65504
  bcs sprite_x_not_out_of_bounds
  brl sprite_out_of_bounds

sprite_not_flipped_horizontally:
  lda a:$0002,x
  clc
  eor a:HFlipMask
  adc ObjectSlot::XPosition
  cmp #256
  bcc sprite_x_not_out_of_bounds
  cmp #65504
  bcs sprite_x_not_out_of_bounds
  brl sprite_out_of_bounds

sprite_x_not_out_of_bounds:
  and #$01FF
  sta a:SpriteBuf3,y		;Store sprite X position SpriteBuf1+y
  and #$00FF
  sta a:SpriteBuf1,y		;Store sprite X position SpriteBuf1+y
  lda a:$0006,x
  and #$0001
  bne sprite_not_flipped_vertically
  sta a:SpriteBuf3+2,y		;sprite size
  lda ObjectSlot::FlipBits
  and #$8000
  beq sprite_not_flipped_vertically
  lda a:$0004,x			;2nd byte = sprite Y position (value 0-255)
  eor a:VFlipMask
  clc
  adc ObjectSlot::YPosition
  clc
  adc #$0008
  cmp #224
  bcc sprite_y_not_out_of_bounds
  cmp #65504
  bcs sprite_y_not_out_of_bounds
  bra sprite_out_of_bounds

sprite_not_flipped_vertically:
  lda a:$0004,x			;2nd byte = sprite Y position (value 0-255)
  eor a:VFlipMask
  clc
  adc ObjectSlot::YPosition
  cmp #224
  bcc sprite_y_not_out_of_bounds
  cmp #65504
  bcs sprite_y_not_out_of_bounds
  bra sprite_out_of_bounds

sprite_y_not_out_of_bounds:
  and #$00FF
  sta a:SpriteBuf1+1,y
  and #$8000
  lda a:$0006,x
  and #$0001
  sta a:SpriteBuf3+2,y		;sprite size
  lda ObjectSlot::FlipBits
  lda a:$0006,x
  and #$0100
  beq sprite_y_not_out_of_bounds_continued
  phx
  ldx a:MetaspriteVramOffset
  lda a:VramLinkedListTable-2,x
  tax
  sta a:MetaspriteVramOffset
  lda a:VramAdressToTileNumberTable-2,x
  sta a:MetaspriteBaseTileNumber
  plx

sprite_y_not_out_of_bounds_continued:
  lda a:$0008,x
  clc
  adc a:MetaspriteBaseTileNumber
  eor ObjectSlot::FlipBits
  sta a:SpriteBuf1+2,y		;extra/character
  iny
  iny
  iny
  iny

sprite_out_of_bounds:
  txa
  clc
  adc #$0008
  tax
  dec a:MetaspriteCount		;decrement MetaspriteCount by 1
  brl metasprite_loop		;back to the loop...

;=========================================================================================

single_metasprite:
  phx
  lda ObjectSlot::VramOffset
  tax
  lda a:VramAdressToTileNumberTable-2,x
  sta a:MetaspriteBaseTileNumber
  plx

  cpy #$0200			;sees if all 128 sprites are used up
  bne single_metasprite_continued
  brl metasprite_finder
  
single_metasprite_continued:
  and #$00FF
  sta a:SpriteBuf1+1,y		;store sprite Y position in SpriteBuf1+1,y

  lda a:$0002,x			;1st byte = sprite X position (value 0-255)
  clc
  adc ObjectSlot::XPosition
  cmp #256
  bcc single_sprite_x_not_out_of_bounds
  cmp #65504
  bcs single_sprite_x_not_out_of_bounds
  bra single_sprite_out_of_bounds

single_sprite_x_not_out_of_bounds:
  and #$01FF
  sta a:SpriteBuf3,y		;store sprite X position SpriteBuf1+y
  and #$00FF
  sta a:SpriteBuf1,y		;store sprite X position SpriteBuf1+y
  lda a:$0004,x			;2st byte = sprite Y position (value 0-255)
  clc
  adc ObjectSlot::YPosition
  cmp #224
  bcc single_sprite_y_not_out_of_bounds
  cmp #65504
  bcs single_sprite_y_not_out_of_bounds
  bra single_sprite_out_of_bounds

single_sprite_y_not_out_of_bounds:
  sta a:SpriteBuf1+1,y
  lda a:$0006,x
  and #$0001
  sta a:SpriteBuf3+2,y		;sprite size
  lda a:$0008,x
  clc
  adc a:MetaspriteBaseTileNumber
  eor ObjectSlot::FlipBits
  sta a:SpriteBuf1+2,y		;extra/character
  iny
  iny
  iny
  iny

single_sprite_out_of_bounds:
  txa
  clc
  adc #$0008
  tax
  dec a:MetaspriteCount		;decrement MetaspriteCount by 1
  brl metasprite_finder
.endproc

;=========================================================================================
;=========================================================================================

tile_uploader:
  rep #$30	; A=16, X/Y=16
  ldx #$0000
  lda #$1801			; Set DMA mode (word, normal increment) and destination register (VRAM write register)
  sta $4300
  sta $4310
  lda #$0080
  sta $2115

tile_uploader_16x16:
  lda TileRequestCounter16x16
  beq tile_uploader_done

  lda #$0040
  sta $4305
  sta $4315

;16x16 Top Half
  lda TileRequestTable+VramAddress,x
  sta $2116

  lda TileRequestTable+BankNumber,x
  sta $4303
  sta $4313

  lda TileRequestTable+TileAddress,x
  sta $4302
  clc
  adc #$0040
  sta $4312

  lda #%0000000000000001	; Initiate DMA transfer (channel 0)
  sta $420B


;16x16 Bottom Half
  lda TileRequestTable+VramAddress,x
  clc
  adc #$0100
  sta $2116

  lda #%0000000000000010	; Initiate DMA transfer (channel 1)
  sta $420B

  txa
  clc
  adc #TileRequestSlotSize
  tax
  cpx TileRequestCounter16x16
  bne tile_uploader_16x16

tile_uploader_done:
  stz TileRequestCounter16x16
  rts

;=========================================================================================
.segment "RODATA"
;=========================================================================================

VramAdressToTileNumberTable:
  .word $0000,$0002,$0004,$0006,$0008,$000A,$000C,$000E
  .word $0020,$0022,$0024,$0026,$0028,$002A,$002C,$002E
  .word $0040,$0042,$0044,$0046,$0048,$004A,$004C,$004E
  .word $0060,$0062,$0064,$0066,$0068,$006A,$006C,$006E
  .word $0080,$0082,$0084,$0086,$0088,$008A,$008C,$008E
  .word $00A0,$00A2,$00A4,$00A6,$00A8,$00AA,$00AC,$00AE
  .word $00C0,$00C2,$00C4,$00C6,$00C8,$00CA,$00CC,$00CE
  .word $00E0,$00E2,$00E4,$00E6,$00E8,$00EA,$00EC,$00EE
  .word $0100,$0102,$0104,$0106,$0108,$010A,$010C,$010E
  .word $0120,$0122,$0124,$0126,$0128,$012A,$012C,$012E
  .word $0140,$0142,$0144,$0146,$0148,$014A,$014C,$014E
  .word $0160,$0162,$0164,$0166,$0168,$016A,$016C,$016E
  .word $0180,$0182,$0184,$0186,$0188,$018A,$018C,$018E
  .word $01A0,$01A2,$01A4,$01A6,$01A8,$01AA,$01AC,$01AE
  .word $01C0,$01C2,$01C4,$01C6,$01C8,$01CA,$01CC,$01CE
  .word $01E0,$01E2,$01E4,$01E6,$01E8,$01EA,$01EC,$01EE
  
VramAddressToTransferAddressTable:
  .word $0000,$0020,$0040,$0060,$0080,$00A0,$00C0,$00E0
  .word $0200,$0220,$0240,$0260,$0280,$02A0,$02C0,$02E0
  .word $0400,$0420,$0440,$0460,$0480,$04A0,$04C0,$04E0
  .word $0600,$0620,$0640,$0660,$0680,$06A0,$06C0,$06E0
  .word $0800,$0820,$0840,$0860,$0880,$08A0,$08C0,$08E0
  .word $0A00,$0A20,$0A40,$0A60,$0A80,$0AA0,$0AC0,$0AE0
  .word $0C00,$0C20,$0C40,$0C60,$0C80,$0CA0,$0CC0,$0CE0
  .word $0E00,$0E20,$0E40,$0E60,$0E80,$0EA0,$0EC0,$0EE0
  .word $1000,$1002,$1004,$1006,$1008,$100A,$100C,$100E
  .word $1200,$1220,$1240,$1260,$1280,$12A0,$12C0,$12E0
  .word $1400,$1420,$1440,$1460,$1480,$14A0,$14C0,$14E0
  .word $1600,$1620,$1640,$1660,$1680,$16A0,$16C0,$16E0
  .word $1800,$1820,$1840,$1860,$1880,$18A0,$18C0,$18E0
  .word $1A00,$1A20,$1A40,$1A60,$1A80,$1AA0,$1AC0,$1AE0
  .word $1C00,$1C20,$1C40,$1C60,$1C80,$1CA0,$1CC0,$1CE0
  .word $1E00,$1E20,$1E40,$1E60,$1E80,$1EA0,$1EC0,$1EE0

;=========================================================================================

TilesInFrameTable:
  .word $0002,$0002

VramAddressFrameTable:
  .word .LOWORD(Test1Tiles),.LOWORD(Test2Tiles)

VramBankByteFrameTable:
  .word .BANKBYTE(Test1Tiles),.BANKBYTE(Test2Tiles)

;=========================================================================================

Test1Tiles:
  .incbin "Test1.pic"

Test2Tiles:
  .incbin "Test2.pic"

;=========================================================================================

TestMetasprite:
  .word $0003	; Number of metasprite table entries below
        ;XPos YPos NextTile/Size Extra/Character
  .word $0000,$0000,$0000,$0001
  .word $0008,$0000,$0001,$0000
  .word $0000,$0010,$0101,$4000

;=========================================================================================

;Byte used to zero/initialise all of memory.
;This is needed because DMA requires an address of a byte to
;read from, not a raw value.

zero_fill_byte:
  .byte $00